﻿using PDFPrintingSample.UWP;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Globalization;
using Windows.Graphics.Display;
using Windows.Graphics.Printing;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI.Xaml.Printing;
using System.Windows;
using Windows.UI.Popups;
using Windows.UI.Core;
using Windows.ApplicationModel.Core;

[assembly: Xamarin.Forms.Dependency(typeof(PrintService))]

namespace PDFPrintingSample.UWP
{
    class PrintService : IPrintService
    {
        internal int m_pageCount;
        internal global::Windows.Data.Pdf.PdfDocument _pdfDocument;
        private PrintDocument printDocument;
        private IPrintDocumentSource printDocumentSource;
        string file_Name;
        internal Dictionary<int, UIElement> printPreviewPages = new Dictionary<int, UIElement>();
        private List<string> m_imagePaths = new List<string>();
        private DispatcherTimer m_dispatcherTimer = new DispatcherTimer();
        IRandomAccessStream randomStream;
        PrintManager printMan;
        double marginWidth;
        double marginHeight;
        Image imageCtrl = new Image();
        PrintTask printTask;
        MessageDialog msgDialog;

        public async void Print(Stream inputStream, string fileName)
        {
            inputStream.Position = 0;
            MemoryStream ms = new MemoryStream();
            inputStream.CopyTo(ms);
            ms.Position = 0;
            randomStream = await ConvertToRandomAccessStream(ms);
            IAsyncOperation<global::Windows.Data.Pdf.PdfDocument> result = null;
            result = global::Windows.Data.Pdf.PdfDocument.LoadFromStreamAsync(randomStream);

            result.AsTask().Wait();
            _pdfDocument = result.GetResults();           
            result = null;
            m_pageCount = (int)_pdfDocument.PageCount;
            file_Name = fileName;
            try
            {
                UIDispatcher.Execute(async () =>
                {
                    RegisterForPrint();
                    await PrintManager.ShowPrintUIAsync();
                });
            }
            catch
            {
                UIDispatcher.Execute(async () =>
                {
                    RegisterForPrint();
                    PrintManager.ShowPrintUIAsync();
                });
            }
        }
        public async Task<IRandomAccessStream> ConvertToRandomAccessStream(MemoryStream memoryStream)
        {
            var randomAccessStream = new InMemoryRandomAccessStream();
            MemoryStream contentStream = new MemoryStream();
            memoryStream.CopyTo(contentStream);
            using (var outputStream = randomAccessStream.GetOutputStreamAt(0))
            {
                using (var dw = new DataWriter(outputStream))
                {
                    var task = new Task(() => dw.WriteBytes(contentStream.ToArray()));
                    task.Start();

                    await task;
                    await dw.StoreAsync();

                    await outputStream.FlushAsync();
                    await dw.FlushAsync();
                    outputStream.Dispose();
                    dw.DetachStream();
                    dw.Dispose();
                }
            }
            return randomAccessStream;
        }
        private void RegisterForPrint()
        {
            printPreviewPages.Clear();
            printDocument = new PrintDocument();
           
            // Save the DocumentSource.
            printDocumentSource = printDocument.DocumentSource;
            // Add an event handler which creates preview pages.
            printDocument.Paginate += CreatePrintPreviewPages;
            // Add an event handler which provides a specified preview page.
            printDocument.GetPreviewPage += GetPrintPreviewPage;
            // Add an event handler which provides all final print pages.
            printDocument.AddPages += AddPrintPages;

            // Create a PrintManager and add a handler for printing initialization.
            printMan = PrintManager.GetForCurrentView();
            printMan.PrintTaskRequested += PrintTaskRequested;
        }
       
        private void AddOnePrintPreviewPage()
        {
            for (int j = 0; j < _pdfDocument.PageCount; j++)
            {
                Canvas canvas = new Canvas();
                canvas.Width = _pdfDocument.GetPage((uint)j).Size.Width;
                canvas.Height = _pdfDocument.GetPage((uint)j).Size.Height;
                printPreviewPages.Add(j, canvas);
            }
        }
        /// <summary>
        /// Sets the specific page as preview.
        /// </summary>
        private void GetPrintPreviewPage(object sender, GetPreviewPageEventArgs e)
        {
#if !SyncfusionFramework4_5
            Canvas print = printPreviewPages[e.PageNumber - 1] as Canvas;

            if (print != null)
            {
                int pageNum = e.PageNumber - 1;
                float zoomfactor = 1;
                double customZoomFactor = zoomfactor * 96 / DisplayProperties.LogicalDpi;
                ApplicationLanguages.PrimaryLanguageOverride = CultureInfo.InvariantCulture.TwoLetterISOLanguageName;
                var pdfPage = _pdfDocument.GetPage(uint.Parse(pageNum.ToString()));
                IRandomAccessStream randomStream = new InMemoryRandomAccessStream();
                global::Windows.Data.Pdf.PdfPageRenderOptions pdfPageRenderOptions = new global::Windows.Data.Pdf.PdfPageRenderOptions();
                Size pdfPageSize = pdfPage.Size;

                pdfPageRenderOptions.DestinationHeight = (uint)(pdfPageSize.Height * customZoomFactor);
                pdfPageRenderOptions.DestinationWidth = (uint)(pdfPageSize.Width * customZoomFactor);
                IAsyncAction actionResult = pdfPage.RenderToStreamAsync(randomStream, pdfPageRenderOptions);
                actionResult.AsTask().Wait();
                Canvas canvas = printPreviewPages[pageNum] as Canvas;
                canvas.Width = (printPreviewPages[pageNum] as Canvas).Width;
                canvas.Height = (printPreviewPages[pageNum] as Canvas).Height;
                imageCtrl = new Image();
                BitmapImage src = new BitmapImage();
                randomStream.Seek(0);
                src.SetSource(randomStream);
                imageCtrl.Source = src;
                imageCtrl.Height = marginHeight;
                imageCtrl.Width = marginWidth;
                if (canvas.Children.Count < 1)
                    canvas.Children.Add(imageCtrl);
                else
                {
                    Image img = canvas.Children[0] as Image;
                    if (img != null)
                    {
                        BitmapImage bitmapImage = img.Source as BitmapImage;
                        bitmapImage.UriSource = null;
                        bitmapImage = null;
                        img = null;
                    }
                    canvas.Children.RemoveAt(0);
                    canvas.Children.Add(imageCtrl);
                }

                pdfPage.Dispose();
                randomStream.Dispose();
                ApplicationLanguages.PrimaryLanguageOverride = CultureInfo.CurrentCulture.TwoLetterISOLanguageName;
            }
#endif
            PrintDocument printDoc = (PrintDocument)sender;
            printDoc.SetPreviewPage(e.PageNumber, printPreviewPages[e.PageNumber - 1]);
        }
        private void CreatePrintPreviewPages(object sender, PaginateEventArgs e)
        {
            PrintTaskOptions printingOptions = ((PrintTaskOptions)e.PrintTaskOptions);
            PrintPageDescription pageDescription = printingOptions.GetPageDescription((uint)e.CurrentPreviewPageNumber);
            marginWidth = pageDescription.PageSize.Width;
            marginHeight = pageDescription.PageSize.Height;
            AddOnePrintPreviewPage();
            PrintDocument printDoc = (PrintDocument)sender;
            printDoc.SetPreviewPageCount(m_pageCount, PreviewPageCountType.Final);
        }
     
        private async void AddPrintPages(object sender, AddPagesEventArgs e)
        {
            try
            {
                await PrepareForPrint(0, m_pageCount);
                PrintDocument printDoc = (PrintDocument)sender;
                printDoc.AddPagesComplete();
            }
            catch
            {
                PrintDocument printDoc = (PrintDocument)sender;
                printDoc.InvalidatePreview();
            }
        }
        private async Task<int> PrepareForPrint(int startIndex, int count)
        {
            StorageFolder tempFolder = ApplicationData.Current.TemporaryFolder;
            int result = await PrepareForPrint(startIndex, count, tempFolder);
            tempFolder = null;
            return result;
        }
        private async Task<int> PrepareForPrint(int p, int count, StorageFolder tempfolder)
        {
            for (int i = p; i < count; i++)
            {
                ApplicationLanguages.PrimaryLanguageOverride = CultureInfo.InvariantCulture.TwoLetterISOLanguageName;
                var pdfPage = _pdfDocument.GetPage(uint.Parse(i.ToString()));
                double pdfPagePreferredZoom = pdfPage.PreferredZoom ;
                IRandomAccessStream randomStream = new InMemoryRandomAccessStream();
                global::Windows.Data.Pdf.PdfPageRenderOptions pdfPageRenderOptions = new global::Windows.Data.Pdf.PdfPageRenderOptions();
                Size pdfPageSize = pdfPage.Size;
                pdfPageRenderOptions.DestinationHeight = (uint)(pdfPageSize.Height * pdfPagePreferredZoom);
                pdfPageRenderOptions.DestinationWidth = (uint)(pdfPageSize.Width * pdfPagePreferredZoom);
                await pdfPage.RenderToStreamAsync(randomStream, pdfPageRenderOptions);
                Image imageCtrl = new Image();
                BitmapImage src = new BitmapImage();
                randomStream.Seek(0);
                src.SetSource(randomStream);
                imageCtrl.Source = src;
                imageCtrl.Height = marginHeight;
                imageCtrl.Width = marginWidth;
                randomStream.Dispose();
                pdfPage.Dispose();
                printDocument.AddPage(imageCtrl);
                if ((printPreviewPages[i] as Canvas).Children.Count > 0)
                    (printPreviewPages[i] as Canvas).Children.RemoveAt(0);
                ApplicationLanguages.PrimaryLanguageOverride = CultureInfo.CurrentCulture.TwoLetterISOLanguageName;             
            }
            return 0;
        }
       
        private async void PrintTaskRequested(PrintManager sender, PrintTaskRequestedEventArgs e)
        {
            printTask = e.Request.CreatePrintTask(file_Name, sourceRequested => sourceRequested.SetSource(printDocumentSource));
            printTask.Completed += printTask_Completed;
         }
        
        private void printTask_Completed(PrintTask sender, PrintTaskCompletedEventArgs args)
        {
            printTask.Completed -= printTask_Completed;
            printMan.PrintTaskRequested -= PrintTaskRequested;
            UIDispatcher.Execute(async () =>
            {
                msgDialog = new MessageDialog("Printing operation has been completed");
                await msgDialog.ShowAsync();                

            });
        }  
    }
    internal class UIDispatcher
    {
        private static CoreDispatcher Dispatcher = CoreApplication.MainView.CoreWindow.Dispatcher;
        /// <summary>
        /// Executes the action using UIElement.
        /// </summary>
        /// <param name="action">An Action.</param>
        internal static void Execute(Action action)
        {
            if (CoreApplication.MainView.CoreWindow == null
                || Dispatcher.HasThreadAccess)
                action();
            else
                Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () => action()).AsTask().Wait();
        }
    }
}
